#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include "ControlCentre.h"
#include "DBElement.h"
#include "Buffer.h"
#include "GnutellaPacket.h"
#include "Connection.h"
#include "Transfer.h"
#include "Download.h"

Download::Download(int newSock, int newFileID, char *newFileName,
		char *newDirName, int newOffset, 
		struct sockaddr_in newsa, char *newCID) 
	: Transfer(newFileID, newFileName, newDirName, newOffset,
			newSock, newsa)
{
	type		= CON_TYPE_DOWNLOAD;
	memcpy(cid, newCID, CLIENTID_LENGTH);
}

Download::Download(Connection *c, int newFileID, char *newFileName,
			char *newDirName, int newOffset, char *newCID)
	: Transfer(newFileID, newFileName, newDirName, newOffset, c)
{
	memcpy(cid, newCID, CLIENTID_LENGTH);
}

Download::Download(Connection *c)
	: Transfer(-1, "", "", 0, c)
{
}

void Download::reInit(int newFileID, char *newFileName, char *newDirName,
				char *newCID)
{
	fileID		= newFileID;
	strncpy(fileName, newFileName, TRANSFER_FN_LENGTH - 2);
	fileName[TRANSFER_FN_LENGTH - 1]	= 0;
	strncpy(dirName, newDirName, TRANSFER_FN_LENGTH - 2);
	dirName[TRANSFER_FN_LENGTH - 1]		= 0;
	memcpy(cid, newCID, CLIENTID_LENGTH);
	type		= CON_TYPE_DOWNLOAD;
}
										
int Download::readBlob()
{
	int	   	bytesRead;
	struct timeval	tm;
	fd_set		in_set;

	tm.tv_sec	= 0;
	tm.tv_usec	= 0;

	FD_ZERO(&in_set);
	FD_SET(connection, &in_set);

	if (select(connection + 1, &in_set, 0, 0, &tm) <= 0)
		return 0;

	bytesRead	= inBuffer->bufRecv(connection);

	if (bytesRead == -1)
	{
		int	i;

		if (bytesLeft <= 0)
			setState(CON_STATE_DONE);
		else
		{
			if (type == CON_TYPE_PUSH_IN)
				setState(CON_STATE_DONE);
			else
				setState(CON_STATE_CON_ERROR);
		}

		/*
		 * at most 5 attempts to write the buffer to file
		 */
		for(i = 0; i < 5; i++)
			if (inBuffer->bytesInBuffer() > 0)
				inBuffer->bufFWrite(file);
			else
				break;

                return 0;
	}

	bytesLeft	-= bytesRead;
	bytesDone	+= bytesRead;
	inBuffer->bufFWrite(file);


	return bytesRead;
}

int Download::waitResponse()
{
	char	getMessage[512];
	int	res;
	char	*str;

	if (getInBufferBytes() <= 0)
		return 0;

	if (((res = peekInBuffer(getMessage, 511))) < 34) // 34 is an estamted minimum size for a http header
		return 0;
	
	getMessage[res]	= 0;
	if (!((str = strstr(getMessage, "\r\n\r\n"))))
		return 0;
	
	if (!strstr(getMessage, "HTTP 200 OK"))
	{
		setState(CON_STATE_LOGIN_FAILED);
		fprintf(stderr, "Error : Download::waitResponse : No HTTP 200 OK !!\n");
		return 0;		
	}
	if ((str = strstr(getMessage, "Content-range: bytes=")))
	{
		int	start;
		int	end;

		str 		+= strlen("Content-range: bytes=");
		sscanf(str, "%d-%d/%d", &start, &end, &bytesTotal);
/*		start		= atoi(str++);
		end		= atoi(str++);
		bytesTotal	= atoi(str);
*/		bytesLeft	= end - start;
		bytesDone	= 0;
	}
	else if ((str = strstr(getMessage, "Content-length:")))
	{
		str 		+= strlen("Content-length:");
		bytesLeft	= bytesTotal	= atoi(str);
		bytesDone	= 0;
	}
	if (!((str = strstr(getMessage, "\r\n\r\n"))))
		return 0;
	str	+= strlen("\r\n\r\n");

	/*
	 * Remove the HTTP header
	 */
	if (res < (int)(str - getMessage))
	{
		fprintf(stderr, "Error : Download : %d > %d!\n",
			(int)(str - getMessage), res);
		
		setState(CON_STATE_LOGIN_FAILED);
		return 0;
	}

	if (skipBytes((int)(str - getMessage)) != (int)(str - getMessage))
	{
		setState(CON_STATE_LOGIN_FAILED);
		fprintf(stderr, "Error : Download : Remove HTTP HEADER failed !!\n");
		return 0;
	}

	setState(CON_STATE_ACTIVE);
	return 1;
}

int Download::loginToPeer()
{
	char	getMessage[512];

	if (offset <= 0)
		sprintf(getMessage, 	"GET /get/%d/%s HTTP/1.0\r\n"
					"Connection: Keep-Alive\r\n\r\n",
					fileID, fileName);
	else
		sprintf(getMessage, 	"GET /get/%d/%s HTTP/1.0\r\n"
					"Connection: Keep-Alive\r\n"
					"Range: bytes=%u-\r\n\r\n",
					fileID, fileName, offset);

	if (!sendStr(getMessage, strlen(getMessage)))
	{
		setState(CON_STATE_LOGIN_FAILED);
		fprintf(stderr, "Error : Download : send Request failed !!!\n");
		return 0;
	}

	setState(CON_STATE_LOGIN_SENT);
	return 1;
}

void Download::getCID(char *buf)
{
	memcpy(buf, cid, CLIENTID_LENGTH);
}

int Download::getIP()
{
	return sa.sin_addr.s_addr;
}

short Download::getPort()
{
	return sa.sin_port;
}

/*
 * FAME
 */

char *Download::getCIDp()
{
	return (char *)cid;
}
