/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <time.h>
#include "ControlCentre.h"
#include "DBElement.h"
#include "GnutellaPacket.h"

GnutellaPacket::GnutellaPacket()
{
	newMesgID();
	header[HEADER_FUNCTION]	= FUNC_INVALID;
	GPinit(0);
	packetBody	= 0;
}

GnutellaPacket::GnutellaPacket(int newCon, ControlCentre *newMCC)
{
	newMesgID();
	header[HEADER_FUNCTION]	= FUNC_INVALID;
	GPinit(newMCC);
	connection	= newCon;
}

GnutellaPacket::GnutellaPacket(ControlCentre *newMCC)
{
	newMesgID();
	header[HEADER_FUNCTION]	= FUNC_INVALID;
	GPinit(newMCC);
}

GnutellaPacket::GnutellaPacket(GnutellaPacket *ask)
{
	int		func;

	ask->getHeader((unsigned char *)header);
	switch (ask->getType())
	{
		case FUNC_INIT		: func = FUNC_INIT_RESP; break;
		case FUNC_SEARCH	: func = FUNC_SEARCH_RESP; break;
//		case FUNC_	: func =; break;
		default		: func = FUNC_INVALID;
	}
	header[HEADER_FUNCTION]		= func;
	*(int *)&header[HEADER_DATALENGTH]	= 0;
	MCC		= ask->getMCC();
	packetBody	= NULL;
	connection	= ask->getConnection();
}

GnutellaPacket::GnutellaPacket(int newPacketType, char *newBody, int len,
				ControlCentre *newMCC)
{
	newMesgID();
	MCC	= newMCC;
	switch (newPacketType)
	{
		case FUNC_INIT		:
				header[HEADER_FUNCTION] = FUNC_INIT;
				break;
		case FUNC_INIT_RESP	:
				header[HEADER_FUNCTION] = FUNC_INIT_RESP;
				break;
		case FUNC_PUSH		:
				header[HEADER_FUNCTION] = FUNC_PUSH;
				break;
		case FUNC_SEARCH	:
				header[HEADER_FUNCTION] = FUNC_SEARCH;
				break;
		case FUNC_SEARCH_RESP	:
				header[HEADER_FUNCTION] = FUNC_SEARCH_RESP;
				break;
/*
		case FUNC_	:
				header[HEADER_FUNCTION] = FUNC_;
				break;
*/
		default 		: header[HEADER_FUNCTION] = FUNC_INVALID;
						break;
	}

	*(int *)&header[HEADER_DATALENGTH]	= len;
	if (newBody)
	{
		packetBody	= (unsigned char *)malloc(len);
		memcpy(packetBody, newBody, len);
	}
	else
	{
		packetBody			= 0;
		*(int *)&header[HEADER_DATALENGTH]	= 0;
	}
}

GnutellaPacket::~GnutellaPacket()
{
	if (packetBody)
		delete [] packetBody;
}

void GnutellaPacket::readHeaderFromString(unsigned char *verdomme)
{
	memcpy(header, verdomme, GNUTELLA_HEADER_LENGTH);

	switch(verdomme[HEADER_FUNCTION])
	{
		case	FUNC_INIT		:
		case	FUNC_INIT_RESP		:
		case	FUNC_PUSH		:
		case	FUNC_SEARCH		:
		case	FUNC_SEARCH_RESP	:
					break;
		default:
				verdomme[HEADER_FUNCTION]	= FUNC_INVALID;
				break;
			
	}
}

void GnutellaPacket::readBodyFromString(unsigned char *buf)
{
	if (packetBody)
		delete [] packetBody;

	if (getDataLength() > 0)
	{
		packetBody	= (unsigned char *)malloc(getDataLength());
		memcpy(packetBody, buf, getDataLength());
	}
	else
		packetBody	= 0;
}

int GnutellaPacket::mkForward()
{
	if ((header[HEADER_HOPS] + header[HEADER_TTL]) > MCC->get_MAX_TTL())
	{
		if (header[HEADER_HOPS] < MCC->get_MAX_TTL())
			header[HEADER_TTL]	= MCC->get_MAX_TTL() - header[HEADER_HOPS];
		else
			return 0;
	}

	if (header[HEADER_TTL] >= MCC->get_MAX_TTL())
		 header[HEADER_TTL]	= MCC->get_MAX_TTL();

	if (header[HEADER_TTL] > 1)
		header[HEADER_TTL]--;

	if (header[HEADER_HOPS] < 255)
		header[HEADER_HOPS]++;

	return (header[HEADER_TTL] &&
			(header[HEADER_HOPS] <= MCC->get_MAX_HOPS()));
}

unsigned int GnutellaPacket::setBody(char *newBody, unsigned int len)
{
	if (packetBody)
		delete [] packetBody;

	*(unsigned int *)&header[HEADER_DATALENGTH]	= len;
	if (newBody)
	{
		packetBody	= (unsigned char *)malloc(len);
		memcpy(packetBody, newBody, len);
	}
	else
	{
		packetBody			= NULL;
		*(unsigned int *)&header[HEADER_DATALENGTH]	= 0;
		if (len)
		{
			MCC->addError("BUG : len without body!!!");
			return 0;
		}
	}
	return len;
}

unsigned int GnutellaPacket::setBodyDirect(char *newBody, unsigned int len)
{
	if (packetBody)
		delete [] packetBody;

	*(unsigned int *)&header[HEADER_DATALENGTH]	= len;
	if ( (!((packetBody	= (unsigned char *)newBody))) && (len > 0) )
		return 0;
	else
		return len;
}

int GnutellaPacket::getBody(unsigned char *buf, unsigned int buflen)
{
	if (*(unsigned int *)&header[HEADER_DATALENGTH] > buflen)
		return -1;

	memcpy(buf, packetBody, *(unsigned int *)&header[HEADER_DATALENGTH]);

	return 	*(int *)&header[HEADER_DATALENGTH];
}

void GnutellaPacket::getHeader(unsigned char *buf)
{
	memcpy(buf, header, GNUTELLA_HEADER_LENGTH);
}

unsigned char GnutellaPacket::getType()
{
	return header[HEADER_FUNCTION];
}

void GnutellaPacket::setType(unsigned char newType)
{
	header[HEADER_FUNCTION]	= newType;
}

void GnutellaPacket::getMesgID(unsigned char *buf)
{
	memcpy((char *)buf, (char *)header, MESSAGE_ID_LENGTH);
}

void GnutellaPacket::setMesgID(unsigned char *buf)
{
	memcpy((char *)header, (char *)buf, MESSAGE_ID_LENGTH);
}

unsigned int GnutellaPacket::getDataLength()
{
	return *(unsigned int *)&header[HEADER_DATALENGTH];
}

int GnutellaPacket::getConnection()
{
	return connection;
}

void GnutellaPacket::setConnection(int newc)
{
	connection	= newc;
}

ControlCentre *GnutellaPacket::getMCC()
{
	return MCC;
}

void GnutellaPacket::sendPacket(int target)
{
	if (target == -1)
	{
		MCC->addError("sendPacket : target == -1!!!");
		return;
	}
	if (getType() == FUNC_INVALID)
	{
		MCC->addError("sendPacket : Type = INVALID!!!");
		return;
	}

	send(target, header, GNUTELLA_HEADER_LENGTH,
#ifdef MSG_NOSIGNAL
		MSG_NOSIGNAL);
#else
		0);
#endif
	if (*(int *)&header[HEADER_DATALENGTH])
	{
		if (packetBody)
			send(target, packetBody, *(unsigned int *)&header[HEADER_DATALENGTH],
#ifdef MSG_NOSIGNAL
				MSG_NOSIGNAL);
#else
				0);
#endif
		else
			fprintf(stderr, "BUG DETECTED in GNUTELLAPACKET!!\n");
	}
}

void GnutellaPacket::GPinit(ControlCentre *newMCC)
{
	MCC				= newMCC;
	header[HEADER_FUNCTION]		= FUNC_INVALID;
	header[HEADER_TTL]		= MCC->get_DEFAULT_TTL();
	*(int *)&header[HEADER_DATALENGTH]	= 0;
	header[HEADER_HOPS]		= 0;
	packetBody			= 0;
}

int GnutellaPacket::isSane()
{
	switch(header[HEADER_FUNCTION])
	{
		case	FUNC_INIT		:
					if (getDataLength())
						return 0;
					if (packetBody)
						return 0;
					return 1;
		case	FUNC_INIT_RESP		:
					if (getDataLength() != INIT_RESP_BODY_SIZE)
						return 0;
					if (!packetBody)
						return 0;
					return 1;
		case	FUNC_PUSH		:
					if (getDataLength() != PUSH_BODY_SIZE)
						return 0;
					if (!packetBody)
						return 0;
					return 1;
		case	FUNC_SEARCH		:
					if (	(getDataLength() < MIN_SEARCH_BODY_SIZE)	||
						(getDataLength() > MAX_SEARCH_BODY_SIZE)	)
						return 0;
					if (!packetBody)
						return 0;
					if (packetBody[getDataLength() - 1])
						return 0;
					return 1;
		case	FUNC_SEARCH_RESP	:
					if (	(getDataLength() < MIN_SEARCH_RESP_BODY_SIZE)	||
						(getDataLength() > MAX_SEARCH_RESP_BODY_SIZE)	)
						return 0;
					if (!packetBody)
						return 0;
					if (!packetBody[0])
						return 0;
					return 1;
		default	: 		return 0;
	}
	return 0;
}

void GnutellaPacket::newMesgID()
{
	unsigned int 	seed;
	unsigned int 	i;

	sprintf((char *)header, "%d", (int)time(0));
	seed	= header[2] * header[3];
	srand(seed);
	i	= rand();
	memcpy(header + 8, (char *)&i, 4);
	i	= rand();
	memcpy(header + 12, (char *)&i, 4);
}

/*
 * DBElement inheritance
 */

int GnutellaPacket::equals(DBElement *gp)
{
	if (connection != ((GnutellaPacket *)gp)->getConnection())
		return 0;

	return (!strncmp((char *)header, (char *)((GnutellaPacket *)gp)->getMesgIDp(),
			 MESSAGE_ID_LENGTH));
}

int GnutellaPacket::lessThan(DBElement *gp)
{
	if (connection > ((GnutellaPacket *)gp)->getConnection())
		return 0;
	if (connection < ((GnutellaPacket *)gp)->getConnection())
		return 1;

	return (strncmp((char *)header, (char *)((GnutellaPacket *)gp)->getMesgIDp(),
			 MESSAGE_ID_LENGTH) < 0);
}

int GnutellaPacket::greaterThan(DBElement *gp)
{
	if (connection < ((GnutellaPacket *)gp)->getConnection())
		return 0;
	if (connection > ((GnutellaPacket *)gp)->getConnection())
		return 1;

	return (strncmp((char *)header, (char *)((GnutellaPacket *)gp)->getMesgIDp(),
			 MESSAGE_ID_LENGTH) > 0);
}

unsigned int GnutellaPacket::hashValue()
{
	return (*(unsigned int *)(header + 5)) & MAX_HASH;
}

/*
 * FAME
 */

char *GnutellaPacket::getBodyp()
{
	return (char *)packetBody;
}

char *GnutellaPacket::getHeaderp()
{
	return (char *)header;
}

char *GnutellaPacket::getMesgIDp()
{
	return (char *)header;
}

