/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
 
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include "Config.h"
#include "ControlCentre.h"
#include "DBElement.h"
#include "DB.h"
#include "Buffer.h"
#include "GnutellaPacket.h"
#include "Connection.h"
#include "GnutellaConnection.h"
#include "Mesg.h"
#include "Host.h"
#include "GnutellaRouter.h"

GnutellaRouter::GnutellaRouter(ControlCentre *newMCC)
{
	MCC		= newMCC;
	messages	= new DB(MCC->get_ROUTING_TABLE_SIZE());
}

GnutellaRouter::~GnutellaRouter()
{
	messages->clearDB();
	delete(messages);
}

void GnutellaRouter::handle(GnutellaPacket *pkt, DB *connections)
{
	if (!pkt)
		return;

	if (!messages)
	{
		MCC->addError("handler() messages test FAILED\n");
		return;
	}

	switch (pkt->getType())
	{
		case	FUNC_INIT	: handle_init(pkt, connections);
						break;
		case	FUNC_INIT_RESP	: handle_init_resp(pkt, connections);
						break;
		case	FUNC_PUSH	: handle_push(pkt, connections);
						break;
		case	FUNC_SEARCH	: handle_search(pkt, connections);
						break;
		case	FUNC_SEARCH_RESP: handle_search_resp(pkt, connections);
						break;
		default			: break;
	}
}

void GnutellaRouter::handle_init(GnutellaPacket *pkt, DB *connections)
{
	handle_even(pkt, connections);
}

void GnutellaRouter::handle_init_resp(GnutellaPacket *pkt,
						DB *connections)
{
	handle_odd(pkt, connections);
}

void GnutellaRouter::handle_push(GnutellaPacket *pkt,
						DB *connections)
{
	handle_even(pkt, connections);
}

void GnutellaRouter::handle_search(GnutellaPacket *pkt,
						DB *connections)
{
	handle_even(pkt, connections);
}

void GnutellaRouter::handle_search_resp(GnutellaPacket *pkt,
						DB *connections)
{
	handle_odd(pkt, connections);
}

void GnutellaRouter::handle_even(GnutellaPacket *pkt,
						DB *connections)
{
	GnutellaConnection	*c;
	Mesg			*msg;
	int			others;

	msg	= new Mesg(pkt);
	if (messages->storeElement(msg) == ENTRY_INVALID)
	{
		delete(msg);
		return;
	}

	if (!pkt->mkForward())
		return;

	for (others = 0; others < connections->getSize(); others++)
	{
		int		con;

		if (others == pkt->getConnection())
			continue;
		c	= (GnutellaConnection *)connections->findElement(others);
		if (c)
			if ( (con = c->getConnection()) != -1)
				if (c->getState() == CON_STATE_ACTIVE)
				{
					c->sendPacket(pkt);
					c->punchSent();
				}
	}
}

void GnutellaRouter::handle_odd(GnutellaPacket *pkt,
						DB *connections)
{
	Mesg			*msg;
	int			i;
	GnutellaConnection	*c;

	msg	= new Mesg(pkt);
	if ((i = messages->containsElement(msg)) == ENTRY_INVALID)
	{
		delete msg;
		if ((c  = (GnutellaConnection *)
                        connections->findElement(pkt->getConnection())))
			c->punchDropped();
		return;
	}

	delete msg;
	msg	= (Mesg *)messages->findElement(i);

	if (pkt->mkForward())
	{
		int		con;

		if ((c	= (GnutellaConnection *)
			connections->findElement(msg->getConnection())))
		{
			if (((con = c->getConnection())) != -1)
				if (c->getState() == CON_STATE_ACTIVE)
				{	
					c->sendPacket(pkt);
					c->punchSent();
				}
		}
		else
			messages->deleteElement(i);
	}
}

int GnutellaRouter::broadcast(GnutellaPacket *pkt,
						DB *connections)
{
	int 	i;
	int 	con;
	int	hostCount;

	if (!pkt)
		return 0;

	hostCount	= 0;

	for (i = 0; i < connections->getSize(); i++)
	{
		GnutellaConnection	*c;

		c	= (GnutellaConnection *)connections->findElement(i);
		if (!c)
			continue;
		if (((con = c->getConnection())) != -1)
			if (c->getState() == CON_STATE_ACTIVE)
			{
				c->sendPacket(pkt);
				hostCount++;
				c->punchSent();
			}
	}

	return hostCount;
}

int GnutellaRouter::alreadySeen(Mesg *m)
{
	return (messages->containsElement(m) != ENTRY_INVALID);
}

int GnutellaRouter::sendPacket(GnutellaPacket *pkt,
				DB *connections)
{
	GnutellaConnection	*c;

	if (!((c = (GnutellaConnection *)
			connections->findElement(pkt->getConnection()))))
		return 0;

	if (c->getState() == CON_STATE_ACTIVE)
	{
		if (c->sendPacket(pkt))
		{
			c->punchSent();
			return 1;
		}
		return 0;
	}

	return 0;
}

int GnutellaRouter::resize()
{
	DB	*newMesgs;
	int	i;

	newMesgs	= new DB(MCC->get_ROUTING_TABLE_SIZE());

	for (i = 0; ( 	(i < newMesgs->getSize())	&&
			(i < messages->getSize())	); i++)
		newMesgs->storeElement(messages->findElement(i));

	while(i < messages->getSize())
		if (messages->findElement(i))
			delete messages->findElement(i);

	delete messages;
	messages	= newMesgs;

	return 1;
}
