/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <unistd.h>
#include <string.h> 
#include <stdio.h>
#include <iostream.h>

#include "ControlCentre.h"
#include "DBElement.h"
#include "Buffer.h"
#include "GnutellaPacket.h"
#include "Connection.h"
#include "Host.h"

Host::Host()
{
	int		i;
	char		dotted[DOTTED_LENGTH];
	struct in_addr	ia;

	for (i = 0; i < 4; i++)
		ip[i]	= 0;
	port		= 0;
	fileCount	= 0;
	fileSize	= 0;
	real		= 1;
	sprintf(dotted, "%hhu.%hhu.%hhu.%hhu", ip[0], ip[1], ip[2], ip[3]);
	inet_aton(dotted, &ia);
	netIP	= ia.s_addr;
	for (i = 0; i < CLIENTID_LENGTH; i++)
		clientID[i]	= 0;
}

Host::Host(unsigned char a, unsigned char b, unsigned char c, unsigned char d,
			unsigned short newPort)
{
	int		i;
	char		dotted[DOTTED_LENGTH];
	struct in_addr	ia;

	ip[0]		= a;
	ip[1]		= b;
	ip[2]		= c;
	ip[3]		= d;
	port		= newPort;
	fileCount	= 0;
	fileSize	= 0;
	real		= 1;
	tries		= 0;

	sprintf(dotted, "%hhu.%hhu.%hhu.%hhu", ip[0], ip[1], ip[2], ip[3]);
	inet_aton(dotted, &ia);
	netIP	= ia.s_addr;
	for (i = 0; i < CLIENTID_LENGTH; i++)
		clientID[i]	= 0;
}

Host::Host(unsigned char a, unsigned char b, unsigned char c, unsigned char d,
			unsigned short newPort, char *cid)
{
	int		i;
	char		dotted[DOTTED_LENGTH];
	struct in_addr	ia;

	ip[0]		= a;
	ip[1]		= b;
	ip[2]		= c;
	ip[3]		= d;
	port		= newPort;
	fileCount	= 0;
	fileSize	= 0;
	real		= 1;
	tries		= 0;
	sprintf(dotted, "%hhu.%hhu.%hhu.%hhu", ip[0], ip[1], ip[2], ip[3]);
	inet_aton(dotted, &ia);
	netIP	= ia.s_addr;
	if (cid)
		memcpy(clientID, cid, CLIENTID_LENGTH);
	else
		for (i = 0; i < CLIENTID_LENGTH; i++)
			clientID[i]	= 0;
}

Host::Host(unsigned char a, unsigned char b, unsigned char c, unsigned char d,
			unsigned short newPort, int fc, int fs)
{
        int		i;
	char		dotted[DOTTED_LENGTH];
	struct in_addr	ia;

        ip[0]           = a;
	ip[1] 		= b;
        ip[2]           = c;
        ip[3]           = d;
        port            = newPort;
	fileCount 	= fc;
        fileSize        = fs;
	real		= 1;
	tries		= 0;
	sprintf(dotted, "%hhu.%hhu.%hhu.%hhu", ip[0], ip[1], ip[2], ip[3]);
	inet_aton(dotted, &ia);
	netIP	= ia.s_addr;
	for (i = 0; i < CLIENTID_LENGTH; i++)
		clientID[i]     = 0;
}

Host::Host(unsigned long newNetIP, unsigned short newPort)
{
	int		i;
	struct in_addr	ia;
	unsigned int	a, b, c, d;

	netIP		= newNetIP;
	ia.s_addr	= newNetIP;
	port		= newPort;
	sscanf(inet_ntoa(ia), "%u.%u.%u.%u", &a, &b, &c, &d);
	ip[0]		= a & 0xFF;
	ip[1]		= b & 0xFF;
	ip[2]		= c & 0xFF;
	ip[3]		= d & 0xFF;
	fileCount	= 0;
	fileSize	= 0;
	real		= 1;
	tries		= 0;
	for (i = 0; i < CLIENTID_LENGTH; i++)
		clientID[i]	= 0;
}

Host::Host(unsigned long newNetIP, unsigned short newPort, int fc, int fs)
{
	int		i;
	struct in_addr	ia;
	unsigned int	a, b, c, d;

	netIP		= newNetIP;
	ia.s_addr	= newNetIP;
	port		= newPort;
	sscanf(inet_ntoa(ia), "%u.%u.%u.%u", &a, &b, &c, &d);
	ip[0]		= a & 0xFF;
	ip[1]		= b & 0xFF;
	ip[2]		= c & 0xFF;
	ip[3]		= d & 0xFF;
 	fileCount	= fc;
	fileSize	= fs;
	real		= 1;
	tries		= 0;
	for (i = 0; i < CLIENTID_LENGTH; i++)
		clientID[i]	= 0;
}

Host::Host(unsigned long newNetIP, unsigned short newPort, char *cid)
{
	struct in_addr	ia;
	unsigned int	a, b, c, d;

	netIP		= newNetIP;
	ia.s_addr	= newNetIP;
	port		= newPort;
	sscanf(inet_ntoa(ia), "%u.%u.%u.%u", &a, &b, &c, &d);
	ip[0]		= a & 0xFF;
	ip[1]		= b & 0xFF;
	ip[2]		= c & 0xFF;
	ip[3]		= d & 0xFF;
	fileCount	= 0;
	fileSize	= 0;
	real		= 1;
	tries		= 0;
	memcpy(clientID, cid, CLIENTID_LENGTH);
}

Host::Host(unsigned long newNetIP, unsigned short newPort, char *cid,
			int fc,int fs)
{
	struct in_addr	ia;
	unsigned int	a, b, c, d;

	netIP		= newNetIP;
	ia.s_addr	= newNetIP;
	port		= newPort;
	sscanf(inet_ntoa(ia), "%u.%u.%u.%u", &a, &b, &c, &d);
	ip[0]		= a & 0xFF;
	ip[1]		= b & 0xFF;
	ip[2]		= c & 0xFF;
	ip[3]		= d & 0xFF;
	port		= newPort;
 	fileCount	= fc;
	fileSize	= fs;
	real		= 1;
	tries		= 0;
	memcpy(clientID, cid, CLIENTID_LENGTH);
}

Host::Host(char *dotted)
{
	int		i;
	struct in_addr	ia;
	unsigned int	a, b, c, d;

	sscanf(dotted, "%u.%u.%u.%u:%hu", &a, &b, &c, &d, &port);
	ip[0]		= a & 0xFF;
	ip[1]		= b & 0xFF;
	ip[2]		= c & 0xFF;
	ip[3]		= d & 0xFF;
 	fileCount	= 0;
	fileSize	= 0;
	real		= 1;
	tries		= 0;
	if (strchr(dotted, ':'))
	{
		*strchr(dotted, ':')	= 0;
		inet_aton(dotted, &ia);
		netIP	= ia.s_addr;
	}
	else
		netIP	= 0;
	netIP	= ia.s_addr;
	for (i = 0; i < CLIENTID_LENGTH; i++)
		clientID[i]	= 0;
}

Host::Host(char *dotted, int fc, int fs)
{
	int		i;
	struct in_addr	ia;
	unsigned int	a, b, c, d;

	sscanf(dotted, "%u.%u.%u.%u:%hu", &a, &b, &c, &d, &port);
	ip[0]		= a & 0xFF;
	ip[1]		= b & 0xFF;
	ip[2]		= c & 0xFF;
	ip[3]		= d & 0xFF;
	fileCount	= fc;
	fileSize	= fs;
	real		= 1;
	tries		= 0;
	if (strchr(dotted, ':'))
	{
		*strchr(dotted, ':')	= 0;
		inet_aton(dotted, &ia);
		netIP	= ia.s_addr;
	}
	else
		netIP	= 0;
	for (i = 0; i < CLIENTID_LENGTH; i++)
		clientID[i]	= 0;
}

Host::Host(char *dotted, char *cid)
{
	int		i;
	struct in_addr	ia;
	unsigned int	a, b, c, d;

	sscanf(dotted, "%u.%u.%u.%u:%hu", &a, &b, &c, &d, &port);
	ip[0]		= a & 0xFF;
	ip[1]		= b & 0xFF;
	ip[2]		= c & 0xFF;
	ip[3]		= d & 0xFF;
	real		= 1;
	tries		= 0;
	if (strchr(dotted, ':'))
	{
		*strchr(dotted, ':')	= 0;
		inet_aton(dotted, &ia);
		netIP	= ia.s_addr;
	}
	else
		netIP	= 0;
	if (cid)
		memcpy(clientID, cid, CLIENTID_LENGTH);
	else
		for (i = 0; i < CLIENTID_LENGTH; i++)
			clientID[i]	= 0;
}

Host::~Host()
{
}

int Host::hasClientID()
{
	int	i;

	for (i = 0; i < CLIENTID_LENGTH; i++)
		if (clientID[i])
			return 1;

	return 0;
}

void Host::incTries()
{
	tries++;
}

void Host::setTries(int t)
{
	tries	= t;
}

int Host::getTries()
{
	return tries;
}

void Host::getClientID(char *buf)
{
	memcpy(buf, clientID, CLIENTID_LENGTH);
}

void Host::setClientID(char *buf)
{
	memcpy(clientID, buf, CLIENTID_LENGTH);
}

int Host::isReal()
{
	return real;
}

void Host::setReal(int nr)
{
	real	= nr;
}

unsigned char Host::getIP(int i)
{
	if ( (i < 0) || (i > 3) )
		return 0;
	else
		return ip[i];
}

unsigned long Host::getNetIP()
{
	return netIP;
}

unsigned short Host::getPort()
{
	return port;
}

Connection *Host::HConnect()
{
	int 			sock;
	struct sockaddr_in      sa;
	Connection		*cn;

        sa.sin_family   	= PF_INET;
        sa.sin_port     	= htons(port);
	sa.sin_addr.s_addr	= netIP;

         if ((sock = socket(PF_INET, SOCK_STREAM, 0)) < 0)
 	{
		fprintf(stderr, "Error : OPENING SOCKET failed!!!\n");
		return 0;
	}

	cn	= new Connection(sock, sa);
	cn->setHostIndex(getEntry());
	cn->setOutgoing();
	cn->tryToConnect();

	return cn;
}

void Host::getDotted(char *buf)
{
	sprintf(buf, "%u.%u.%u.%u:%u", ip[0], ip[1], ip[2], ip[3], port);
}

/*
 * DBElement Inheritance
 */

int Host::equals(DBElement *myHost)
{
	if (netIP != ((Host *)myHost)->getNetIP())
		return 0;
	if (port != (((Host *)myHost))->getPort())
		return 0;

	return 1;
}

int Host::lessThan(DBElement *myHost)
{
	if (netIP < ((Host *)myHost)->getNetIP())
		return 1;
	if (netIP > ((Host *)myHost)->getNetIP())
		return 0;

	if (port < ((Host *)myHost)->getPort())
		return 1;
	if (port > ((Host *)myHost)->getPort())
		return 0;

	return 0;
}

int Host::greaterThan(DBElement *myHost)
{
	if (netIP > ((Host *)myHost)->getNetIP())
		return 1;
	if (netIP < ((Host *)myHost)->getNetIP())
		return 0;

	if (port > ((Host *)myHost)->getPort())
		return 1;
	if (port < ((Host *)myHost)->getPort())
		return 0;

	return 0;
}

unsigned int Host::hashValue()
{
	return netIP & MAX_HASH;
}

/*
 * FAME
 */

char *Host::getClientIDp()
{
	return (char *)clientID;
}
